import os
import mimetypes


# http versions:
HTTP_VERSION_1_0 = 'HTTP/1.0'
HTTP_VERSION_1_1 = 'HTTP/1.1'

# http response statuses and messages:
HTTP_CONTINUE = [100, 'Continue']
HTTP_SWITCHING_PROTOCOLS = [101, 'Switching Protocols']
HTTP_OK = [200, 'OK']
HTTP_NO_CONTENT = [204, 'No Content']

HTTP_MOVED_PERMANENTLY = [301, 'Moved Permanently']
HTTP_NOT_MODIFIED = [304, 'Not Modified']

HTTP_BAD_REQUEST = [400, 'Bad Request']
HTTP_UNAUTHORIZED = [401, 'Unauthorized']
HTTP_FORBIDDEN = [403, 'Forbidden']
HTTP_NOT_FOUND = [404, 'Not Found']
HTTP_METHOD_NOT_ALLOWED = [405, 'Method Not Allowed']
HTTP_REQUEST_TIMEOUT = [408, 'Request Time-out']
HTTP_LENGTH_REQUIRED = [411, 'Length Required']
HTTP_URI_TOO_LONG = [414, 'Request-URI Too Large']
HTTP_REQUEST_HEADER_FIELDS_TOO_LARGE = [431, 'Request Header Fields Too Large']

HTTP_INTERNAL_SERVER_ERROR = [500, 'Internal Server Error']
HTTP_NOT_IMPLEMENTED = [501, 'Not Implemented']
HTTP_SERVICE_UNAVAILABLE = [503, 'Service Unavailable']
HTTP_VERSION_NOT_SUPPORTED = [505, 'HTTP Version not supported']


class HttpException(Exception):
    def __init__(self, error, request=None):
        self.error = error
        self.request = request


class Headers:
    def __init__(self):
        self._headers = dict()

    def add_header(self, name, value):
        self._headers[name.strip().lower()] = value

    def remove_header(self, name):
        if self.has_header(name):
            del self._headers[name.lower()]

    def __getattr__(self, name):
        if self.has_header(name):
            return self.get_header(name)

        return ''

    def get_header(self, name):
        return self._headers.get(name.lower())

    def get_header_line(self, name):
        return '{name}: {value}'.format(name=name.title(), value=self.get_header(name))

    def has_header(self, name, expected_value=None):
        value = self.get_header(name)

        if expected_value is not None:
            return value == expected_value

        return value is not None


class Request(Headers):
    def __init__(self, method, uri, http_version):
        super(Request, self).__init__()

        self.method = method.strip().upper()
        self.uri = uri
        self.http_version = http_version.strip().upper()
        self.body = None


class Response(Headers):
    def __init__(self, status, body=None):
        super(Response, self).__init__()

        self.status_code, self.reason_phrase = status
        self._body = None
        self.http_version = HTTP_VERSION_1_1

        self.body = body

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, body):
        self._body = body

        # https://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.4
        if self._should_not_have_body():
            self._body = None

        if self._body is None:
            self.remove_header('Content-Length')
        else:
            self.add_header('Content-Length', len(self._body))

    def stream(self):
        yield bytes(str(self), 'utf-8')

    def __str__(self):
        formatted_headers = '\r\n'
        if len(self._headers) > 0:
            headers = []
            for name in self._headers:
                headers.append(self.get_header_line(name))
            headers.append('\r\n')
            formatted_headers = '\r\n'.join(headers)

        return ('{http_version} {status_code} {reason_phrase}\r\n' +
                '{headers}' +
                '{body}') \
            .format(http_version=self.http_version,
                    status_code=str(self.status_code),
                    reason_phrase=str(self.reason_phrase),
                    headers=formatted_headers,
                    body=self.body or '')

    def _should_not_have_body(self):
        return (HTTP_CONTINUE[0] <= self.status_code < HTTP_OK[0]) or \
               self.status_code in (HTTP_NO_CONTENT[0], HTTP_NOT_MODIFIED[0])


class FileResponse(Response):
    def __init__(self, path):
        super(FileResponse, self).__init__(HTTP_OK, None)
        self.path = path
        self.file_size = os.path.getsize(self.path)
        self.add_header('Content-Length', self.file_size)

        content_type = mimetypes.guess_type(self.path)[0]
        if content_type:
            self.add_header('Content-Type', content_type)

    def stream(self, buffer_length=8192):
        for bytes_ in super(FileResponse, self).stream():
            yield bytes_

        with open(self.path, 'rb') as content_file:
            while True:
                bytes_ = content_file.read(buffer_length)
                if bytes_:
                    yield bytes_
                else:
                    return
