import socket
from collections import deque

from tak.middleware import HttpErrorHandler, Middleware
from tak.http import HTTP_VERSION_1_0, HttpException, HTTP_BAD_REQUEST, HTTP_REQUEST_TIMEOUT, \
    HTTP_INTERNAL_SERVER_ERROR, HTTP_NOT_FOUND, Request
from tak.parser import RequestReader, ParseException, Parser


class ConnectionClosedException(Exception):
    pass


class SocketRequestReader(RequestReader):
    def __init__(self, socket_, buffer_length, logger):
        self.socket = socket_
        self.buffer_length = buffer_length
        self.logger = logger

    def read_request(self):
        request = bytearray()

        while True:
            new_data = self.socket.recv(self.buffer_length)
            if not new_data:
                raise ConnectionClosedException()

            request = request + new_data

            if self._exists_end_chars(request):
                break

        self.logger.debug('Client sent bytes: %s', request)

        return request

    def read_body(self, length):
        # @todo handle Content-length

        return ''

    def _exists_end_chars(self, request):
        return request.find(b'\n\n') != -1 or request.find(b'\r\n\r\n') != -1


class HttpHandler:
    def __init__(self, socket_, address, server_logger, logger_, error_handler, request_reader, parser_, middleware_):
        self.socket = socket_
        self.ip = address[0]
        self.server_logger = server_logger
        self.logger = logger_
        self.error_handler = error_handler
        self.request_reader = request_reader
        self.parser = parser_
        self.middleware = middleware_

    def handle(self):
        self.server_logger.debug('HttpHandler started')

        with self.socket:
            connection_closed = False
            while not connection_closed:
                try:
                    request, response = self._handle_exchange()
                except HttpException as e:
                    request = e.request
                    response = self.error_handler.handle_error(e.error)

                if request is None:
                    connection_closed = True
                else:
                    if request.has_header('Connection', 'close'):
                        connection_closed = True

                    if request.http_version == HTTP_VERSION_1_0 and not request.has_header('Connection', 'keep-alive'):
                        connection_closed = True

                if response is not None:
                    if connection_closed:
                        response.add_header('Connection', 'close')

                    try:
                        response_length = self._send_response(response)
                    except ConnectionResetError:
                        self.logger.debug('Connection Reset error')
                        connection_closed = True
                        response_length = '-'
                    self._log(request, response, response_length)

        self.server_logger.debug('HttpHandler stopped')

    def _log(self, request, response, response_length):
        context = {
            'ip': self.ip,
            'request': Request('-', '-', '-'),
            'response': response,
            'response_length': response_length,
        }

        if request is not None:
            context['request'] = request

        self.logger.info('', extra=context)

    def _handle_exchange(self):
        try:
            request = self.parser.parse(self.request_reader)
            self.server_logger.debug('Request parsed: %s', request)

            response = self.middleware.handle_request(request)
            self.server_logger.debug('Middleware returned response: %s', bytearray(str(response), 'utf-8'))
            if not response:
                raise HttpException(HTTP_NOT_FOUND, request)

        except ParseException as e:
            self.server_logger.debug('Parsing error: %s', e)
            raise HttpException(HTTP_BAD_REQUEST)
        except socket.timeout as e:
            self.server_logger.debug('Socket reading timeout')
            raise HttpException(HTTP_REQUEST_TIMEOUT)
        except ConnectionClosedException as e:
            self.server_logger.debug('Connection closed by client')
            request = response = None
        except HttpException as e:
            raise e
        except Exception as e:
            self.server_logger.debug('Unexpected error')
            raise HttpException(HTTP_INTERNAL_SERVER_ERROR)

        return request, response

    def _send_response(self, response):
        length = 0
        for bytes_ in response.stream():
            self.socket.send(bytes_)
            length = length + len(bytes_)
        return length


class HttpHandlerModule:
    def create_default_http_handler(self, config, connection, address, server_logger):
        error_handler = HttpErrorHandler(config)
        request_reader = SocketRequestReader(connection, config['read_buffer_length'], server_logger)
        parser_ = Parser()
        middleware_ = Middleware(config, server_logger)
        return HttpHandler(connection, address, server_logger, config['logger_http'],
                           error_handler, request_reader, parser_, middleware_)

    def process_config(self, config):
        if 'middleware' not in config:
            config['middleware'] = deque()

        if 'read_buffer_length' not in config:
            config['read_buffer_length'] = 1024
        config['read_buffer_length'] = int(config['read_buffer_length'])

        config['http_handler'] = self.create_default_http_handler
