from tak.http import Request, HTTP_VERSION_1_1


class ParseException(Exception):
    pass


class RequestReader:
    def read_request(self):
        raise NotImplementedError

    def read_body(self, length):
        raise NotImplementedError


class StringRequestReader(RequestReader):
    def __init__(self, string_request, string_body=''):
        self.request = string_request
        self.body = string_body

    def read_request(self):
        return self.request

    def read_body(self, length):
        return self.body


class Parser:
    REQUEST_LINE_SEPARATORS = (' ', '\t')
    HEADER_KEY_VALUE_SEPARATORS = (':',)
    BODY_SEPARATORS = ('\r\n\r\n', '\n\n')
    LINE_SEPARATORS = ('\r', '\n')
    WHITESPACE_CHARS = (' ', '\t')

    def __init__(self, request=b''):
        self.request = str(request, encoding='utf-8')
        self.parsed_request = None
        self.cursor = 0

    def parse(self, request_reader):
        try:
            self.request = str(request_reader.read_request(), encoding='utf-8')
        except UnicodeDecodeError:
            raise ParseException('Unicode decode error')

        self.cursor = 0

        if self.is_request_empty():
            raise ParseException('Empty request')

        method, uri, http_version = self._parse_request_line()
        self.parsed_request = Request(method, uri, http_version)

        self._parse_headers()

        if self.parsed_request.http_version == HTTP_VERSION_1_1 and self.parsed_request.has_header('Host') is False:
            raise ParseException('Missing Host header')

        return self.parsed_request

    def _parse_request_line(self):
        method = self.request_line_element()
        uri = self.request_line_element()
        http_version = self.request_line_http_version()

        self.skip_end_line()

        return method, uri, http_version

    def _parse_headers(self):
        while True:
            if self.has_request_ended():
                raise ParseException('Invalid request')

            next_char = self.next_char(increment=False)
            if next_char in self.LINE_SEPARATORS:
                return

            name = self.header_name()
            self.skip(self.HEADER_KEY_VALUE_SEPARATORS)

            value = self.header_value()
            self.skip_end_line()

            self.parsed_request.add_header(name, value)

    def request_line_element(self):
        return self.slice_to(self.REQUEST_LINE_SEPARATORS, invalid_chars=self.LINE_SEPARATORS)

    def request_line_http_version(self):
        return self.slice_to(self.LINE_SEPARATORS)

    def header_name(self):
        return self.slice_to(self.HEADER_KEY_VALUE_SEPARATORS, invalid_chars=self.LINE_SEPARATORS).lower().capitalize()

    def header_value(self):
        return self.slice_to(self.LINE_SEPARATORS)

    def slice_to(self, end_chars, invalid_chars=()):
        self.skip(self.WHITESPACE_CHARS)

        start_index = self.cursor

        while True:
            if self.has_request_ended():
                raise ParseException('Invalid request')

            current_char = self.next_char()
            if current_char in invalid_chars:
                raise ParseException('Invalid request')

            if current_char in end_chars:
                break

        self.prev_char()

        return self.request[start_index:self.cursor].strip()

    def next_char(self, increment=True):
        char = self.request[self.cursor]
        if increment:
            self.cursor += 1

        return char

    def prev_char(self):
        char = self.request[self.cursor]
        self.cursor -= 1

        return char

    def skip_end_line(self):
        self.skip(self.WHITESPACE_CHARS)

        end_line_char = self.next_char()
        if end_line_char == '\n':
            return

        if end_line_char == '\r':
            end_line_char = self.next_char()
            if end_line_char == '\n':
                return

        raise ParseException('Missing end line')

    def skip(self, chars):
        while not self.has_request_ended():
            if self.request[self.cursor] not in chars:
                break

            self.cursor += 1

    def has_request_ended(self):
        return self.cursor >= len(self.request)

    def is_request_empty(self):
        return len(self.request) == 0
