import os
import pkg_resources

from tak.config import ConfigException
from tak.http import FileResponse, HTTP_FORBIDDEN, HttpException


class ServeFilesApp:
    def __init__(self, config):
        self.base_uri = config['base_uri']
        self.root_dir = config['root_dir']

    def handle(self, request):
        if request.uri.find(self.base_uri) != 0:
            return None

        path = self.root_dir + request.uri.replace(self.base_uri, '/')

        if os.path.isfile(path):
            if os.access(path, os.R_OK):
                return FileResponse(path)
            raise HttpException(HTTP_FORBIDDEN)

        return None


class ServeIconsApp(ServeFilesApp):
    def __init__(self, config):
        self.base_uri = config['_icon_base_uri']
        self.root_dir = config['_icon_root_dir']


class AutoIndexApp:
    def __init__(self, config):
        self.base_uri = config['base_uri']
        self.root_dir = config['root_dir']
        self.view = config['view']
        self.icon_base_uri = config['_icon_base_uri']

    def handle(self, request):
        if request.uri.find(self.base_uri) != 0:
            return None

        path = self.root_dir + request.uri.replace(self.base_uri, '/')
        if not os.path.isdir(path):
            return None

        current_uri = request.uri
        if not request.uri.endswith('/'):
            current_uri = request.uri + '/'

        show_up = current_uri != self.base_uri
        __, dirs, files = next(os.walk(path))

        dirs.sort()
        files.sort()

        return self.view().render_response('servefiles/listing', {
            'dirs': dirs,
            'files': files,
            'base_uri': current_uri,
            'show_up': show_up,
            'icon_prefix': self.icon_base_uri
        })


class ServeFilesModule:
    def process_config(self, config):
        if 'servefile' not in config:
            config['servefile'] = 'on'

        if config['servefile'] not in ['on', 'off']:
            raise ConfigException('Invalid value for "servefile". Valid options: on|off')

        if 'autoindex' not in config:
            config['autoindex'] = 'on'

        if config['autoindex'] not in ['on', 'off']:
            raise ConfigException('Invalid value for "autoindex". Valid options: on|off')

        if 'base_uri' not in config:
            config['base_uri'] = '/'

        if 'root_dir' not in config:
            config['root_dir'] = './'

        if not os.path.isdir(config['root_dir']):
            raise ConfigException('Directory for "root_dir" not exists: "{root_dir}"'.format(root_dir=config['root_dir']))

        config['_icon_base_uri'] = '/e7058c8274df4e828868db73e06731e3/'
        config['_icon_root_dir'] = pkg_resources.resource_filename('tak', 'templates/servefiles/icons/')

        if config['servefile'] == 'on':
            config['middleware'].append(ServeFilesApp)

        if config['autoindex'] == 'on':
            config['middleware'].append(ServeIconsApp)
            config['middleware'].append(AutoIndexApp)
