import signal
import socket
import threading

from tak.log import get_server_logger
from tak.config import ConfigException, load_from_ini_file


class Server:
    def __init__(self, config_path, default_modules, logger=None):
        self.config_path = config_path
        self.default_modules = default_modules
        self.logger = logger or get_server_logger()
        self.config = None
        self.socket = None
        self.transport = None

    def start(self):
        try:
            signal.signal(signal.SIGINT, self._handle_sig_int)

            self._do_start()
        except ConfigException:
            self.logger.exception("Config error")
        except OSError as e:
            self.logger.exception("OSError occured")
        except Exception as e:
            self.logger.exception("Unexpected error occured")
            self.stop()

    def _do_start(self):
        self.config = load_from_ini_file(self.config_path, self.default_modules)
        self.config.process()

        with self.config['transport'].create_socket() as self.socket:
            self.logger.info("Server started: {ip}:{port}".format(ip=self.config['host'], port=self.config['port']))

            request_counter = 0
            while True:
                connection, addr = self.socket.accept()
                connection.settimeout(self.config['read_timeout'])
                request_counter += 1

                handler = self.config['http_handler'](self.config, connection, addr, self.logger)
                thread = threading.Thread(target=handler.handle, daemon=True)
                thread.start()

    def _handle_sig_int(self, sig, unused):
        self.stop()

    def stop(self):
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
            self.socket.close()

            self.logger.info("Server stopped: {ip}:{port}".format(ip=self.config['host'],
                                                                  port=self.config['port']))
        except Exception as e:
            self.logger.exception('Exception during server stop')


class ServerModule:
    def process_config(self, config):
        if 'host' not in config:
            config['host'] = '127.0.0.1'

        if 'port' not in config:
            config['port'] = 8000

        try:
            config['port'] = int(config['port'])
        except ValueError:
            raise ConfigException('Invalid value for "port": "{value}"'.format(value=config['port']))

        if 'read_timeout' not in config:
            config['read_timeout'] = 60

        config['read_timeout'] = int(config['read_timeout'])
